@@ -1,32 +1,69 @@
-import subprocess, os, struct, cStringIO, collections, re
-import hglib, error, util, templates, merge
+import subprocess, os, struct, cStringIO, re, datetime
+import hglib, error, util, templates, merge, context
 
 from util import cmdbuilder
 
+class revision(tuple):
+    def __new__(cls, rev, node, tags, branch, author, desc, date):
+        return tuple.__new__(cls, (rev, node, tags, branch, author, desc, date))
+
+    @property
+    def rev(self):
+        return self[0]
+
+    @property
+    def node(self):
+        return self[1]
+
+    @property
+    def tags(self):
+        return self[2]
+
+    @property
+    def branch(self):
+        return self[3]
+
+    @property
+    def author(self):
+        return self[4]
+
+    @property
+    def desc(self):
+        return self[5]
+
+    @property
+    def date(self):
+        return self[6]
+
 class hgclient(object):
     inputfmt = '>I'
     outputfmt = '>cI'
     outputfmtsize = struct.calcsize(outputfmt)
     retfmt = '>i'
 
-    revision = collections.namedtuple('revision', 'rev, node, tags, '
-                                                  'branch, author, desc')
-
-    def __init__(self, path, encoding, configs):
-        args = [hglib.HGPATH, 'serve', '--cmdserver', 'pipe',
+    def __init__(self, path, encoding, configs, connect=True):
+        self._args = [hglib.HGPATH, 'serve', '--cmdserver', 'pipe',
                 '--config', 'ui.interactive=True']
         if path:
-            args += ['-R', path]
+            self._args += ['-R', path]
         if configs:
-            args += ['--config'] + configs
-        env = {}
+            self._args += ['--config'] + configs
+        self._env = {}
         if encoding:
-            env['HGENCODING'] = encoding
+            self._env['HGENCODING'] = encoding
 
-        self.server = util.popen(args, env)
-        self._readhello()
+        self.server = None
         self._version = None
 
+        if connect:
+            self.open()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
+
     def _readhello(self):
         """ read the hello message the server sends when started """
         ch, msg = self._readchannel()
@@ -59,10 +96,17 @@ class hgclient(object):
         else:
             return channel, self.server.stdout.read(length)
 
-    def _parserevs(self, splitted):
+    @staticmethod
+    def _parserevs(splitted):
         ''' splitted is a list of fields according to our rev.style, where each 6
         fields compose one revision. '''
-        return [self.revision._make(rev) for rev in util.grouper(6, splitted)]
+        revs = []
+        for rev in util.grouper(7, splitted):
+            # truncate the timezone and convert to a local datetime
+            posixtime = float(rev[6].split('.', 1)[0])
+            dt = datetime.datetime.fromtimestamp(posixtime)
+            revs.append(revision(rev[0], rev[1], rev[2], rev[3], rev[4], rev[5], dt))
+        return revs
 
     def runcommand(self, args, inchannels, outchannels):
         def writeblock(data):
@@ -133,6 +177,14 @@ class hgclient(object):
                 return eh(ret, out, err)
         return out
 
+    def open(self):
+        if self.server is not None:
+            raise ValueError('server already open')
+
+        self.server = util.popen(self._args, self._env)
+        self._readhello()
+        return self
+
     def close(self):
         """
         Closes the command server instance and waits for it to exit, returns the
@@ -163,7 +215,8 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('add', *files, n=dryrun, S=subrepos, I=include, X=exclude)
+        args = cmdbuilder('add', n=dryrun, S=subrepos, I=include, X=exclude,
+                          *files)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -195,8 +248,8 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('addremove', *files, s=similarity, n=dryrun, I=include,
-                          X=exclude)
+        args = cmdbuilder('addremove', s=similarity, n=dryrun, I=include,
+                          X=exclude, *files)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -227,9 +280,9 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('annotate', *files, r=rev, no_follow=nofollow, a=text,
-                          u=user, f=file, d=date, n=number, c=changeset, l=line,
-                          v=verbose, I=include, X=exclude)
+        args = cmdbuilder('annotate', r=rev, no_follow=nofollow, a=text,
+                          u=user, f=file, d=date, n=number, c=changeset,
+                          l=line, v=verbose, I=include, X=exclude, *files)
 
         out = self.rawcommand(args)
 
@@ -437,7 +490,7 @@ class hgclient(object):
         "%d"  dirname of file being printed, or '.' if in repository root
         "%p"  root-relative path name of file being printed
         """
-        args = cmdbuilder('cat', *files, r=rev, o=output)
+        args = cmdbuilder('cat', r=rev, o=output, *files)
         out = self.rawcommand(args)
 
         if not output:
@@ -501,7 +554,7 @@ class hgclient(object):
         if not isinstance(names, list):
             names = [names]
 
-        args = cmdbuilder('showconfig', *names, u=untrusted, debug=showsource)
+        args = cmdbuilder('showconfig', u=untrusted, debug=showsource, *names)
         out = self.rawcommand(args)
 
         conf = []
@@ -551,8 +604,8 @@ class hgclient(object):
             source = [source]
 
         source.append(dest)
-        args = cmdbuilder('copy', *source, A=after, f=force, n=dryrun,
-                          I=include, X=exclude)
+        args = cmdbuilder('copy', A=after, f=force, n=dryrun,
+                          I=include, X=exclude, *source)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -585,12 +638,12 @@ class hgclient(object):
         if change and revs:
             raise ValueError('cannot specify both change and rev')
 
-        args = cmdbuilder('diff', *files, r=revs, c=change,
+        args = cmdbuilder('diff', r=revs, c=change,
                           a=text, g=git, nodates=nodates,
                           p=showfunction, reverse=reverse,
                           w=ignoreallspace, b=ignorespacechange,
                           B=ignoreblanklines, U=unified, stat=stat,
-                          S=subrepos, I=include, X=exclude)
+                          S=subrepos, I=include, X=exclude, *files)
 
         return self.rawcommand(args)
 
@@ -619,8 +672,8 @@ class hgclient(object):
         """
         if not isinstance(revs, list):
             revs = [revs]
-        args = cmdbuilder('export', *revs, o=output, switch_parent=switchparent,
-                          a=text, g=git, nodates=nodates)
+        args = cmdbuilder('export', o=output, switch_parent=switchparent,
+                          a=text, g=git, nodates=nodates, *revs)
 
         out = self.rawcommand(args)
 
@@ -643,7 +696,7 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('forget', *files, I=include, X=exclude)
+        args = cmdbuilder('forget', I=include, X=exclude, *files)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -677,9 +730,9 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('grep', *[pattern] + files, all=all, a=text, f=follow,
-                          i=ignorecase, l=fileswithmatches, n=line, u=user, d=date,
-                          I=include, X=exclude)
+        args = cmdbuilder('grep', all=all, a=text, f=follow, i=ignorecase,
+                          l=fileswithmatches, n=line, u=user, d=date,
+                          I=include, X=exclude, *[pattern] + files)
         args.append('-0')
 
         def eh(ret, out, err):
@@ -719,8 +772,8 @@ class hgclient(object):
         if not isinstance(rev, list):
             rev = [rev]
 
-        args = cmdbuilder('heads', *rev, r=startrev, t=topological, c=closed,
-                          template=templates.changeset)
+        args = cmdbuilder('heads', r=startrev, t=topological, c=closed,
+                          template=templates.changeset, *rev)
 
         def eh(ret, out, err):
             if ret != 1:
@@ -791,10 +844,11 @@ class hgclient(object):
             prompt = None
             input = None
 
-        args = cmdbuilder('import', *patches, strip=strip, force=force,
-                          nocommit=nocommit, bypass=bypass, exact=exact,
-                          importbranch=importbranch, message=message,
-                          date=date, user=user, similarity=similarity, _=stdin)
+        args = cmdbuilder('import', strip=strip, force=force,
+                          no_commit=nocommit, bypass=bypass, exact=exact,
+                          import_branch=importbranch, message=message,
+                          date=date, user=user, similarity=similarity, _=stdin,
+                          *patches)
 
         self.rawcommand(args, prompt=prompt, input=input)
 
@@ -887,11 +941,11 @@ class hgclient(object):
         include - include names matching the given patterns
         exclude - exclude names matching the given patterns
         """
-        args = cmdbuilder('log', *files, template=templates.changeset,
+        args = cmdbuilder('log', template=templates.changeset,
                           r=revrange, f=follow, follow_first=followfirst,
                           d=date, C=copies, k=keyword, removed=removed,
                           m=onlymerges, u=user, b=branch, P=prune, h=hidden,
-                          l=limit, M=nomerges, I=include, X=exclude)
+                          l=limit, M=nomerges, I=include, X=exclude, *files)
 
         out = self.rawcommand(args)
         out = out.split('\0')[:-1]
@@ -980,8 +1034,8 @@ class hgclient(object):
             source = [source]
 
         source.append(dest)
-        args = cmdbuilder('move', *source, A=after, f=force, n=dryrun,
-                          I=include, X=exclude)
+        args = cmdbuilder('move', A=after, f=force, n=dryrun,
+                          I=include, X=exclude, *source)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -1152,7 +1206,8 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('remove', *files, A=after, f=force, I=include, X=exclude)
+        args = cmdbuilder('remove', A=after, f=force, I=include, X=exclude,
+                          *files)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -1179,8 +1234,8 @@ class hgclient(object):
         if not isinstance(file, list):
             file = [file]
 
-        args = cmdbuilder('resolve', *file, a=all, l=listfiles, m=mark, u=unmark,
-                          t=tool, I=include, X=exclude)
+        args = cmdbuilder('resolve', a=all, l=listfiles, m=mark, u=unmark,
+                          t=tool, I=include, X=exclude, *file)
 
         out = self.rawcommand(args)
 
@@ -1220,8 +1275,9 @@ class hgclient(object):
         if not isinstance(files, list):
             files = [files]
 
-        args = cmdbuilder('revert', *files, r=rev, a=all, d=date,
-                          no_backup=nobackup, n=dryrun, I=include, X=exclude)
+        args = cmdbuilder('revert', r=rev, a=all, d=date,
+                          no_backup=nobackup, n=dryrun, I=include, X=exclude,
+                          *files)
 
         eh = util.reterrorhandler(args)
         self.rawcommand(args, eh=eh)
@@ -1309,8 +1365,8 @@ class hgclient(object):
         if not isinstance(names, list):
             names = [names]
 
-        args = cmdbuilder('tag', *names, r=rev, m=message, f=force, l=local,
-                          remove=remove, d=date, u=user)
+        args = cmdbuilder('tag', r=rev, m=message, f=force, l=local,
+                          remove=remove, d=date, u=user, *names)
 
         self.rawcommand(args)
 
@@ -1475,3 +1531,6 @@ class hgclient(object):
             self._version = tuple(v)
 
         return self._version
+
+    def __getitem__(self, changeid):
+        return context.changectx(self, changeid)
